### /u/sy/beebe/tex/bib/usmarc-to-bibtex.awk, Mon Nov 15 10:26:43 2004
### Edit by Nelson H. F. Beebe <beebe@math.utah.edu>
### ====================================================================
### Convert USMARC library catalog data to (rough) BibTeX form.  The
### definition of USMARC fields can be found on the World-Wide Web at
###
###	http://www.loc.gov/marc/bibliographic/ecbdhome.html
###
### Usage:
###	awk -f usmarc-to-bibtex.awk MARC-file(s) >BibTeX-file
###
### A typical entry retrieved by ISBN looks like this:
###
###	035    $9 (DLC)   86021423
###	906    $a 7 $b cbc $c orignew $d 1 $e ocip $f 19 $g y-gencatlg
###	010    $a    86021423
###	020    $a 0070431523 (pbk.) : $c $19.95
###	040    $a DLC $c DLC $d DLC
###	050 00 $a QA76.76.O63 $b M68 1987
###	082 00 $a 005.4/465 $2 19
###	100 1  $a Morgan, Rachel.
###	245 10 $a Introducing UNIX System V / $c Rachel Morgan and Henry McGilton.
###	260    $a New York : $b McGraw-Hill, $c c1987.
###	300    $a xxv, 612 p. : $b ill. ; $c 23 cm.
###	500    $a Includes index.
###	630 00 $a UNIX System V (Computer file)
###	700 1  $a McGilton, Henry.
###	991    $b c-GenColl $h QA76.76.O63 $i M68 1987 $p 00034708459 $t Copy 1 $w BOOKS
###
### The same entry retrieved by the second author looks like this:
###
###	001 3751753
###	005 19910522102432.4
###	008 821022s1983    nyua          001 0 eng
###	035    $9 (DLC)   82021640
###	906    $a 7 $b cbc $c orignew $d 1 $e ocip $f 19 $g y-gencatlg
###	010    $a    82021640
###	020    $a 0070450013 (pbk.) : $c $18.95
###	040    $a DLC $c DLC $d DLC
###	050 00 $a QA76.8.U65 $b M38 1983
###	082 00 $a 001.64/4 $2 19
###	100 1  $a McGilton, Henry.
###	245 10 $a Introducing the UNIX system / $c Henry McGilton, Rachel Morgan.
###	260    $a New York : $b McGraw-Hill, $c c1983.
###	300    $a xix, 556 p. : $b ill. ; $c 23 cm.
###	440  0 $a McGraw-Hill software series for computer professionals
###	440  0 $a Byte books
###	504    $a Bibliography: p. 541-544.
###	500    $a Includes index.
###	630 00 $a UNIX (Computer file)
###	700 1  $a Morgan, Rachel.
###	991    $b c-GenColl $h QA76.8.U65 $i M38 1983 $p 00003514328 $t Copy 1 $w BOOKS
###
### Notice that the 100 and 700 records do not correctly identify the
### primary and secondary authors in the search by second author; only
### the 245.$c record does.
###
### [25-Mar-2005] -- add support for MARC 090 and extend fix_LCCN()
### [15-Nov-2004] -- original version
### ====================================================================

BEGIN		{ initialize() }

		{ gsub("[\r]+$", "") }	# reduce CR LF to LF

		{ gsub("\r", "") }

/^[ \t]*$/	{ print_entry(); next }

/^nextResultSetPosition/	{ print_entry(); next }

$0 ~ "Z> +open +"		{ Bibsource = $3 }

$1 == "001"	{ print_entry(); next }

$1 == "020"	{ USMARC_020($0); next }
$1 == "021"	{ USMARC_021($0); next }
$1 == "022"	{ USMARC_022($0); next }
$1 == "030"	{ USMARC_030($0); next }
$1 == "050"	{ USMARC_050($0); next }
$1 == "090"	{ USMARC_090($0); next }
$1 == "097"	{ USMARC_097($0); next }
$1 == "100"	{ USMARC_100($0); next }
$1 == "110"	{ USMARC_110($0); next }
$1 == "111"	{ USMARC_111($0); next }
$1 == "130"	{ USMARC_130($0); next }
$1 == "245"	{ USMARC_245($0); next }
$1 == "250"	{ USMARC_250($0); next }
$1 == "260"	{ USMARC_260($0); next }
$1 == "300"	{ USMARC_300($0); next }
$1 == "440"	{ USMARC_440($0); next }
$1 == "490"	{ USMARC_490($0); next }
$1 == "500"	{ USMARC_500($0); next }
$1 == "505"	{ USMARC_505($0); next }
$1 == "520"	{ USMARC_520($0); next }
$1 == "600"	{ USMARC_600($0); next }
$1 == "630"	{ USMARC_630($0); next }
$1 == "650"	{ USMARC_650($0); next }
$1 == "651"	{ USMARC_651($0); next }
$1 == "690"	{ USMARC_690($0); next }
$1 == "700"	{ USMARC_700($0); next }
$1 == "740"	{ USMARC_740($0); next }
$1 == "745"	{ USMARC_745($0); next }
$1 == "776"	{ USMARC_776($0); next }
$1 == "852"	{ USMARC_852($0); next }
$1 == "856"	{ USMARC_856($0); next }

### We explicitly ignore these uninteresting keys:

$1 == "005"	{ next }
$1 == "008"	{ next }
$1 == "010"	{ next }
$1 == "015"	{ next }
$1 == "035"	{ next }
$1 == "040"	{ next }
$1 == "041"	{ next }
$1 == "042"	{ next }
$1 == "082"	{ next }
$1 == "240"	{ next }
$1 == "504"	{ next }
$1 == "800"	{ next }
$1 == "906"	{ next }
$1 == "920"	{ next }
$1 == "922"	{ next }
$1 == "925"	{ next }
$1 == "955"	{ next }
$1 == "991"	{ next }

($0 ~ "^[ \t]")	{   # we have a continuation of the last USMARC record
		    if ($0 ~ "^[ \t]+$")
			;	# ignore blank lines
		    else if (Last_Key in Suffix) # continuation lines are always joined by space
		    {
			sub(/[.;] *$/,"",Entry[Last_Key]) # remove old separator
			append(Last_Key, trim($0), " ")
		    }
		    else
		    {
			unrecognized($1,$0 " DEBUG: Last_Key = [" Last_Key "]")
			Last_Key = ""
		    }
		    next
		}

		{ unrecognized($1,$0); Last_Key = ""; next }

END		{ terminate() }

# ========================================================================


function append(key,v,separator)
{
    if (key in Entry)
	set(key,Entry[key] separator v)
    else
	set(key,v)
}


function append_suffix(key,v)
{
    append(key, v, Suffix[key])
    Last_Key = key		# global value need to handle continuation lines
}


function author_editor()
{
    if ("author" in Entry)
	return (Entry["author"])
    else if ("editor" in Entry)
	return (Entry["editor"])
    else
    {
	Entry["author"] = "Anonymous"
	return (Entry["author"])
    }
}


function capitalize(s, k,n,parts)
{
    if (is_corporate_name(s))
	return (s)		# leave corporate names alone
    if (s ~ /[A-Z][A-Z]/)	# Reduce HEARN, A. C. to Hearn, A. C.
    {				# but leave McMahon, L. E. unchanged
	n = split(s,parts," ")
	for (k = 1; k <= n; ++k)
	    s = (k == 1) ? \
		capitalize_word(parts[k]) : \
		s " " capitalize_word(parts[k])
    }
    return (s)
}


function capitalize_word(s, k)
{
    for (k = 1; k <= length(s); ++k)
    {
	## if (isupper(substr(s,k,1)) && (k > 1) && isletter(substr(s,k-1,1)))
	## [14-May-1996]: Change isletter() above to isupper(), so that
	## we do not convert "McLean" to "Mclean".
	if (isupper(substr(s,k,1)) && (k > 1) && isupper(substr(s,k-1,1)))
	    s = substr(s,1,k-1) tolower(substr(s,k,1)) substr(s,k+1)
    }
    return (s)
}


function check_CODEN(the_CODEN,    checksum)
{
    ## CODEN values are 6-character strings from the set [A-Z0-9],
    ## with a check digit stored in the 6th position given by
    ##
    ##     (11*N1 + 7*N2 + 5*N3 + 3*N4 + 1*N5) mod 34 == X
    ##
    ## where the Nk are 1..26 for A..Z, and 27..36 for 1..9..0.
    ## However, the checksum X (in 0..33) is represented by the
    ## corresponding character in the different 34-character range
    ## [9A-Z2-8], which excludes digits 0 and 1 to avoid confusion
    ## with letters O and I.

    checksum = 0
    checksum += 11 * CODEN_Digit(the_CODEN, 1)
    checksum +=  7 * CODEN_Digit(the_CODEN, 2)
    checksum +=  5 * CODEN_Digit(the_CODEN, 3)
    checksum +=  3 * CODEN_Digit(the_CODEN, 4)
    checksum +=  1 * CODEN_Digit(the_CODEN, 5)

    if (int(checksum % 34) != CODEN_Check_Digit(the_CODEN, 6))
	warning("invalid CODEN: [" the_CODEN "]")
}


function check_CODEN_list(CODEN_list,    k,n,parts)
{
    CODEN_list = toupper(CODEN_list)
    n = split(CODEN_list, parts, "[,; ]+")
    for (k = 1; k <= n; ++k)
	check_CODEN(parts[k])
}


function check_ISBN(the_ISBN,    checksum,isbn,k)
{
    ## ISBN-10 values are strings containing nine characters from
    ## [0-9] and a tenth check digit in [0-9X], ignoring hyphens and spaces.
    ## The check digits is valid if
    ##	(sum(k=1:9) digit(k) * k) mod 11 == digit(10)
    ## where X as a digit has value 10.

    isbn = the_ISBN
    gsub("[- ]", "", isbn)

    checksum = 0
    for (k = 1; k <= 9; ++k)
	checksum += ISBN_Digit(isbn, k) * k

    if (int(checksum % 11) != ISBN_Digit(isbn, 10))
	warning("invalid ISBN: [" the_ISBN "]")
}


function check_ISBN_list(ISBN_list,    k,n,parts)
{
    ISBN_list = toupper(ISBN_list)
    n = split(ISBN_list, parts, "[,; ]+")
    for (k = 1; k <= n; ++k)
    {
	if (parts[k] !~ "^([^()]*)$")	# ignored parenthesized comments
	    check_ISBN(parts[k])
    }
}


function check_ISBN_13(the_ISBN_13,    checkdigit,checksum,isbn,k,remainder)
{
    ## The conversion from ISBN-10 to ISBN-13 is the same process
    ## currently used to encode the ISBN-10 into the 13-digit
    ## Bookland EAN:
    ##
    ## Step 1: Drop the check digit (the last digit) from your
    ## existing ISBN-10.
    ##
    ## For example - Your ISBN-10 is: 0-940016-61-3
    ## Dropping the check digit gives you a 9-digit number:
    ## 0-940016-61
    ##
    ## Step 2: Add the prefix "978" to the beginning of your 9-digit number.
    ##
    ## So, 0-940016-61 becomes 978-0-940016-61
    ##
    ## Step 3: Recalculate your check digit using the modulus 10
    ## check digit routine. This is the current routine used to
    ## calculate the check digit for the Bookland EAN.
    ##
    ##     EAN =                   9    7   8   0    9    4    0   0   1    6    6    1
    ##     Weighting Factors       1    3   1   3    1    3    1   3   1    3    1    3
    ##     Values (product)        9 + 21 + 8 + 0 +  9 + 12 +  0 + 0 + 1 + 18 +  6 +  3 +
    ##
    ## Sum of the products=87
    ## Divide the sum by the modulus (always 10)
    ## Quotient=10 with a remainder of 7
    ## The modulus (10) minus the remainder equals the check digit,
    ## with one exception.  If the remainder from Step 2 is 0, the
    ## check digit is 0.  In this case, however: 10-7=3.
    ## ISBN-13 check digit = 3
    ##
    ## Your ISBN-13 is: 978-0-940016-61-3

    isbn = the_ISBN_13
    gsub("[- ]", "", isbn)

    checksum = 0
    for (k = 1; k <= 12; ++k)
	checksum += ISBN_Digit(isbn, k) * ((int(k % 2) == 1) ? 1 : 3)

    remainder = int(checksum % 10)
    checkdigit = (remainder == 0) ? 0 : (10 - remainder)
    if (checkdigit != ISBN_Digit(isbn, 13))
	warning("invalid ISBN-13: [" the_ISBN_13 "]")
}


function check_ISBN_13_list(ISBN_13_list,    k,n,parts)
{
    ISBN_13_list = toupper(ISBN_13_list)
    n = split(ISBN_13_list, parts, "[,; ]+")
    for (k = 1; k <= n; ++k)
    {
	if (parts[k] !~ "^([^()]*)$")	# ignored parenthesized comments
	    check_ISBN_13(parts[k])
    }
}


function check_ISSN(the_ISSN,    checksum,issn,k)
{
    ## ISSN numbers are 8-character values from the set [0-9Xx], with
    ## a checksum given by
    ##
    ##	(sum(k=1:7) digit(k) * (k+2)) mod 11 == digit(8)
    ##
    ## where digits have their normal value, X (or x) as a digit has
    ## value 10, and spaces and hyphens are ignored.  The sum is
    ## bounded from above by 10*(3 + 4 + ... + 9) = 420.

    sub("-", "", the_ISSN)
    issn = toupper(the_ISSN)

    checksum = 0
    for (k = 1; k <= 7; ++k)
	checksum += ISSN_Digit(issn, k) * (k + 2)

    if (int(checksum % 11) != ISSN_Digit(issn, 8))
	warning("invalid ISSN: [" the_ISSN "]")
}


function check_ISSN_list(ISSN_list,    k,n,parts)
{
    ISSN_list = toupper(ISSN_list)
    n = split(ISSN_list, parts, "[,; ]+")
    for (k = 1; k <= n; ++k)
    {
	if ( (parts[k] != "(PRINT)") && (parts[k] != "(ELECTRONIC)") )
	    check_ISSN(parts[k])
    }
}


function check_for_proceedings(    s)
{
    if (("title" in Entry) && !("booktitle" in Entry))
    {
	s = tolower(Entry["title"] " " \
		    (("note" in Entry) ? Entry["note"] : "") \
		    (("USMARC-711" in Entry) ? Entry["USMARC-711"] : ""))
	if (s ~ "conference|congress|proceedings|school|symposium|workshop")
	{
	    Entry["booktitle"] = Entry["title"]
	    set("entrytype","Proceedings")
	    if (("author" in Entry) && !("editor" in Entry))
	    {
		Entry["editor"] = Entry["author"]
		delete Entry["author"]
	    }
	}
    }
}


function citation_label( author,abbrev,k,n,parts,t,tag)
{
    author = author_editor()
    if (substr(author,1,1) == "{")
    {				# have corporate author
	split(author,parts," ")	# "{SPSS, Inc."}
	t = substr(parts[1],2)	# "SPSS,"
	gsub(/,/,"",t)		# reduce to "SPSS"
    }
    else
	t = capitalize(author)
    gsub(/ and /, ",", t)		# for uniform field separators
    split(t,parts,",")	# careful: "De La Beaujardiere, Jean-M" has spaces in name
    gsub(/[^-A-Za-z'\055]/,"",parts[1])	# strip non-(letters,apostrophe)

    sub(/Jr$/,"",parts[1])	# remove any Jr suffix
    tag = parts[1] ":" Entry["year"] ":"

    ## We do a second strip here, because some entries for the year
    ## can have spaces and other words in them, and because we do
    ## not want apostrophes in the tag (bibxxx utilities complain).
    gsub(/[^-A-Za-z0-9:\055]/,"",tag)	# strip non-(letters,digits,colon))

    n = split(Entry["title"],parts," ")
    abbrev = ""
    for (k = 1; (k <= n) && (length(abbrev) < 3); ++k)
    {
	gsub(/[^A-Za-z0-9\'\055]/,"",parts[k]) # strip non-(letters,digits,apostrophe)
	if (isletter(substr(parts[k],1,1)))
	{
	    parts[k] = tolower(parts[k])
	    if (!(parts[k] in Ignore))
		abbrev = abbrev toupper(substr(parts[k],1,1))
	}
    }
    return (unique_citation_label(tag abbrev))
}


function clean_author_editor(s)
{
    ## Typical values:
    ## Smith, David, 1948 June 26-
    ## Danuloff, Craig, 1963-
    ## Dayton, Linnea

    gsub(/19[0-9][0-9]-? *([A-Z][a-z]+ [0-9]+-?)/,"",s)
    return (s)
}


function clean_entry()
{
    if ("author" in entry)
	set("author",clean_author_editor(Entry["author"]))

    if ("editor" in Entry)
	set("editor",clean_author_editor(Entry["editor"]))

    if ("title" in Entry)
    {
	gsub(/ *:/,":",Entry["title"])		# remove space before colons
	gsub(/ *\/[^\/]*$/,"",Entry["title"])	# strip author name(s) from title
	gsub("[ ]+[.][ ]+", " ", Entry["title"]) # strip isolated periods preceding subtitles
	gsub("[.]$", "", Entry["title"])	# strip trailing period
	gsub("}-{", "-", Entry["title"])	# e.g., reduce {S}-{PLUS} to {S-PLUS}
	if ("booktitle" in Entry)
	    Entry["booktitle"] = Entry["title"]
    }
    if ("edition" in Entry)
    {
	sub(/ ed[.]?$/,"",Entry["edition"])
	if (Entry["edition"] == "1st")
	    delete Entry["edition"]
	else if (Entry["edition"] == "2nd")
	    set("edition","Second")
	else if (Entry["edition"] == "3rd")
	    set("edition","Third")
	else if (Entry["edition"] == "4th")
	    set("edition","Fourth")
	else if (Entry["edition"] == "5th")
	    set("edition","Fifth")
	else if (Entry["edition"] == "6th")
	    set("edition","Sixth")
	else if (Entry["edition"] == "7th")
	    set("edition","Seventh")
	else if (Entry["edition"] == "8th")
	    set("edition","Eighth")
	else if (Entry["edition"] == "9th")
	    set("edition","Nineth")
	else if (Entry["edition"] == "10th")
	    set("edition","Tenth")
	else if (Entry["edition"] == "Rev.")
	    set("edition","Revised")
    }

    if ("ISBN" in Entry)
    {
	gsub(/ISBN */,"",Entry["ISBN"])
	clean_ISBN()
	Entry["ISBN"] = fix_ISBN(Entry["ISBN"])
    }

    if (("note" in Entry) && (substr(Entry["note"],length(Entry["note"]),1) != "."))
	Entry["note"] = Entry["note"] "."

    if ("price" in Entry)	# NB: must come after ISBN processing
    {
	gsub(/ *U[.]?S[.]?A[.]? */," ",Entry["price"])
	gsub(/ *,/,",",Entry["price"])
    }

    if ("address" in Entry)
	Entry["address"] = fix_address(Entry["address"])

    fix_annote()

    if ("author-dates" in Entry)
    {
	if (Entry["author-dates"] ~ "-- *$") # do not record dates for living authors
		delete Entry["author-dates"]
    }

    if ("note" in Entry)
	Entry["note"] = fix_note(Entry["note"])

    if ("pages" in Entry)
	Entry["pages"] = fix_pages(Entry["pages"])

    Entry["LCCN"] = ("LCCN" in Entry) ? fix_LCCN(Entry["LCCN"]) : "????"

    if ("abstract" in Entry)
	Entry["abstract"] = fix_abstract(Entry["abstract"])

    if ("subject" in Entry)
	Entry["subject"] = fix_subject(Entry["subject"])

    if ("URL" in Entry)
	Entry["URL"] = fix_URL(Entry["URL"])

    if ("volume" in Entry)
	Entry["volume"] = fix_volume(Entry["volume"])

    if ("year" in Entry)
	Entry["year"] = fix_year(Entry["year"])

    if (Bibsource != "")
	Entry["bibsource"] = Bibsource
}


function clean_ISBN( ISBN,parts)
{
    ## Typical values
    ##	"0070240493 (H) : US\$29.95, 0070240507 (P) : US\$19.95"
    ##   "0442019963 (Van Nostrand Reinhold), 1850321418 (ITP UK)"
    ##   "0125184085 (acid-free paper)"
    ##   "0125531303 (alk. paper), 0125531311 (CD-Rom : alk. paper)"
    ##   "0127378405 (alk. paper), 0127378413 (CD-ROM)"
    ##   "0201633957 (alk. paper)"
    ##   "0262112078 (alk. paper), 026261118X (pbk. : alk. paper)"
    ##   "0471038571 (alk. paper : pbk.) : US\$19.95"
    ##   "0838976972 (alk. paper) : US\$18.00 (LITA mbr. US\$16.20)"
    ##   "0887369332 (alk. paper : pbk.) : US\$30.00 (US\$37.50 Can. : L21.00 UK)"
    ##   "0471130885 (cloth : alk. paper)"
    ##   "0841232237 (clothbound), 0841232245 (pbk.)"
    ##   "0078811821 (cover), 0078821821: US\$34.95"
    ##   "0079121004 (p)"
    ##   "0471113360 (paper : acid-free paper) : US\$16.95"
    ##   "0471128430 (paper : alk. paper)"
    ##   "0712332952 (pbk)"
    ##   "0471126942 (pbk. : acid-free paper)"
    ##   "047114178X (pbk. : alk. paper)"
    ##   "007024054X (pbk.) : US\$19.95"
    ##   "020140978X (pbk.) US\$26.95 (US\$34.95 Can.)"
    ##   "2707125024 (pbk.) : 140FF"
    ##   "0192627058 (pbk.)"
    ##   "0131515659 (recycled paper)"
    ##   "0387945547 (softcover : alk. paper)"
    ##   "0446671797 (trade pbk.)"
    ##   "088736943X (v. 1 : alk. paper) : US\$7.00, 0887369448 (v. 2 : alk. paper) : US\$7.00, 0887369456 (v. 3 : alk. paper) : US\$7.00, 0887369464 (v. 4 : alk. paper) : US\$7.00, 0887369472 (v. 5 : alk. paper) : US\$7.00, 0887369480 (v. 6 : alk. paper) : US\$7.00"
    ##   "012553132X, 0125531338 (CD-ROM)"
    ##   "1556158858, 1556158181"
    ##   "0070240558 : US\$16.95"
    ##   "0070670188 : US\$45.00, 0070670196 (pbk.) : US\$29.95"
    ##   "0077079051 : L22.95"
    ##   "013185612X : US\$29.95 (US\$39.95 Can.)"
    ##   "0201410001 : US\$17.95 (Can. US\$22.95)"
    ##   "0201627078 : US\$14.95 (US\$18.95 Can.)"
    ##   "0672305305 : US\$25.00 (US\$34.95 CAN)"
    ##   "1562054376 : US\$35.00 (US\$47.95 Can.) (L32.49 UK)"
    ##   "156761521X : US\$12.99"
    ##   "0201407663 US\$12.95 (US\$16.95 Can.)"
    ##   "0789701375 US\$39.99 (US\$53.99 Can.)"

    ISBN = entry["ISBN"]

    gsub(/[$]/,"US\\$",ISBN)

    gsub(/ +,/,",",ISBN)
    gsub(/\(v[.] /,"(vol. ",ISBN)
    gsub(/acid-free paper/,"",ISBN)
    gsub(/alk[.] paper/,"",ISBN)
    gsub(/mbr[.]/,"members",ISBN)
    gsub(/pbk[.]/,"paperback",ISBN)
    gsub(/pbk/,"paperback",ISBN)
    gsub(/\( *\)/,"",ISBN)
    gsub(/\([pP]\)/,"(paperback)",ISBN)
    gsub(/\([hH]\)/,"(hardcover)",ISBN)
    gsub(/ *: *\)/,")",ISBN)
    gsub(/\( *: */,"(",ISBN)
    gsub(/CD-Rom/,"CD-ROM",ISBN)
    gsub(/C[Aa][Nn][Aa][Dd][Aa] */,"Can.",ISBN)
    gsub(/C[Aa][Nn][.]? *US/,"CDN",ISBN)

    gsub(/[ :]+$/,"",ISBN)
    gsub(/[ ]+/," ",ISBN)

    while (match(ISBN,/US\\[$][0-9.]+ C[Aa][Nn][.]/))
    {
	ISBN = substr(ISBN,1,RSTART-1) "CDN" substr(ISBN,RSTART+2,RLENGTH-7) \
	  substr(ISBN,RSTART+RLENGTH)
    }

    while (match(ISBN,/US\\[$][0-9.]+ C[Aa][Nn]/))
    {
	ISBN = substr(ISBN,1,RSTART-1) "CDN" substr(ISBN,RSTART+2,RLENGTH-6) \
	  substr(ISBN,RSTART+RLENGTH)
    }

    while (match(ISBN,/L[0-9.]+/))
    {
	ISBN = substr(ISBN,1,RSTART-1) "UK\\pounds" substr(ISBN,RSTART+1,RLENGTH-1) \
	  substr(ISBN,RSTART+RLENGTH)
    }

    while (match(ISBN,/[0-9] : UK/))
    {
	ISBN = substr(ISBN,1,RSTART) "," substr(ISBN,RSTART+3)
    }

    while (match(ISBN,/\(CDN\\\$[0-9.]+\)/))
    {
	ISBN = substr(ISBN,1,RSTART-1) ", " substr(ISBN,RSTART+1,RLENGTH-2)
    }

    while (match(ISBN,/\(UK\\pounds[0-9.]+\)/))
    {
	ISBN = substr(ISBN,1,RSTART-1) ", " substr(ISBN,RSTART+1,RLENGTH-2)
    }

    gsub(/[ ]+,/,",",ISBN)
    gsub(/[ :]+$/,"",ISBN)
    gsub(/[ ]+/," ",ISBN)

    if ((index(ISBN,":") == 0) && \
	match(ISBN,/[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9xX],?( \([^)]*\))? *[A-Z]/))
	ISBN = substr(ISBN,1,RSTART + RLENGTH - 2) ": " substr(ISBN,RSTART + RLENGTH - 1)

    if (split(ISBN,parts,":") == 2)
    {
	ISBN = parts[1]

	gsub(/^ +/,"",parts[2])
	gsub(/[ ]+/," ",parts[2])
	entry["price"] = entry["price"] ", " parts[2]
	sub(/^, */,"",entry["price"])
    }

    gsub(/[ ,]+$/,"",ISBN)
    entry["ISBN"] = ISBN
}


function clear_array(array, key)
{
    for (key in array)
	delete array[key]
}


function CODEN_Check_Digit(the_CODEN,n)
{
    return (index("9ABCDEFGHIJKLMNOPQRSTUVWXYZ2345678", substr(the_CODEN, n, 1)) - 1)
}


function CODEN_Digit(the_CODEN,n)
{
    return (index("ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890", substr(the_CODEN, n, 1)))
}


function delete_match(s)
{
    ## After a match() call on s, return s after deleting the match
    return (substr(s,1,RSTART-1) substr(s,RSTART+RLENGTH))
}


function enquote(v)
{
    if (substr(v,1,1) == "\"")	# enquote leading quote
	v = "{\377}" substr(v,2)
    while (match(v,/[^\\]\"/))
    {				# brace unbackslashed quotes
	v = substr(v,1,RSTART) "{\377}" substr(v,RSTART+2)
    }
    gsub(/\377/,"\"",v)

    gsub(/&/," and ",v)		# replace ampersands
    gsub(/[ \t][ \t]+/," ",v)	# eliminate redundant spaces
    gsub(/[%]/,"\\%",v)		# protect quotes
    gsub(/[#]/,"\\#",v)		# protect sharps
    gsub(/[_]/,"\\_",v)		# protect underscores
    gsub(/[\^]/,"\\^",v)	# protect carets

    if (v ~ /[0-9][$][0-9]/)
	gsub(/[$]/,"--",v)	# June 7$10 -> June 7--10
#    else if (v ~ /[$][0-9]+/)
#	gsub(/[$]/,"US\\$",v)	# $27 -> US\$27 and $27.95 -> US\$27.95
    return (v)
}


function entry_name( )
{
    return (("entrytype" in Entry) ? \
	Entry["entrytype"] : "????UNKNOWN-TYPE-????")
}


function fix_abstract(s)
{
    ## Remove duplicate MARC-record joining dots (from Suffix["abstract"])
    gsub("[.][.][.]* ", ". ", s)
    sub("[.][.][.]*$",  ".",  s)

    return (s)
}


function fix_address(s)
{
    s = trim(squeeze_blanks(s))
    ## print "%% DEBUG: enter fix_address(\"" s "\")"

    sub(" *[:;/] *$", "", s)

    ## Major publishers are in only a few states
    sub(" Cal[.]$",	" CA, USA",	s)
    sub(" Calif[.]$",	" CA, USA",	s)
    sub(" Mass[.]$",	" MA, USA",	s)
    sub(" Minn[.]$",	" MN, USA",	s)
    sub(" N[.]J[.]$",	" NJ, USA",	s)
    sub(" N[.]Y[.]$",	" NY, USA",	s)
    sub("Washington, D[.]C[.]$",	"Washington, DC, USA",	s)

    ## Major publishers are in only a few cities
    sub("^Boston$",		"Boston, MA, USA",	s)
    sub("^Chicago$",		"Chicago, IL, USA",	s)
    sub("^New York$",		"New York, NY, USA",	s)
    sub("^Philadelphia$",	"Philadelphia, PA, USA",s)
    sub("^San Francisco$",	"San Francisco, CA, USA",	s)
    sub("^Washington$",		"Washington, DC, USA",	s)

    ## Fix a few non-US publisher addresses
    sub("^Amsterdam$",	"Amsterdam, The Netherlands",	s)
    sub("^Berlin$",	"Berlin, Germany",	s)
    sub("^Brussels$",	"Brussels, Belgium",	s)
    sub("^Bruxelles$",	"Bruxelles, Belgium",	s)
    sub("^Copenhagen$",	"Copenhagen, Denmark",	s)
    sub("^Edinburgh$",	"Edinburgh, Scotland",	s)
    sub("^Heidelberg$",	"Heidelberg, Germany",	s)
    sub("^Lisbon$",	"Lisbon, Portugal",	s)
    sub("^London$",	"London, UK",		s)
    sub("^Madrid$",	"Madrid, Spain",	s)
    sub("^Montreal$",	"Montr{\\'e}al, Qu{\\'e}bec, Canada",	s)
    sub("^Montr{\\\\\'e}al$",	"Montr{\\'e}al, Qu{\\'e}bec, Canada",	s)
    sub("^Mnchen$",	"M{\\\"u}nchen, Germany",	s)
    sub("^Oslo$",	"Oslo, Norway",		s)
    sub("^Oxford$",	"Oxford, UK",		s)
    sub("^Paris$",	"Paris, France",	s)
    sub("^Roma$",	"Roma, Italy",		s)
    sub("^Stockholm$",	"Stockholm, Sweden",	s)
    sub("^Stuttgart$",	"Stuttgart, Germany",	s)
    sub("^Toronto$",	"Toronto, ON, Canada",	s)
    sub("^Toronto, Canada$",	"Toronto, ON, Canada",	s)
    sub("^Warszawa$",	"Warszawa, Poland",	s)

    ## print "%% DEBUG: leave fix_address(\"" s "\")"

    return (s)
}


function fix_annote( n,parts,e,s)
{
    ## Remove some unwanted phrases in annote fields.

    if ("annote" in Entry)
    {
	s = Entry["annote"]
	gsub(/Includes bibliograph?ic(al)? references *(and index(es)?)? *(\([^\)]*\))?[.]?/,"",s)
	gsub(/Includes bibliograph?ic(al)? references *(p[.])? *(\([^\)]*\))? *(and index(es)?)?[.]?/,"",s)
	gsub(/Includes bibliograph?ic(al)? references *(and index(es)?)?[.]?/,"",s)
	gsub(/Includes bibliograph?(ies|y) *(and index(es)?)?[.]?/,"",s)
	gsub(/Includes glossar(y|ies) *(and index(es)?)?[.]?/,"",s)
	gsub(/Includes (author)? *index(es)?[.]?/,"",s)
	gsub(/Includes bibliograph?y[.]? *\([^\)]*\)[.]?/,"",s)
	gsub(/Includes bibliograph?(ic|ical|y) references[.]? *(\([^\)]*\))?(, *glossar(ies|y))? *(and *index(es)?)?[.]?/,"",s)
	gsub(/Includes bibliograph?(y|ies) \([^\)]*\) *(and index(es)?)[.]?*/,"",s)

	gsub(/Title from cover[.]?/,"",s)
	gsub(/Bibliograph?(y|ies) *: *(leaves|pages|pp[.]?|p[.]?) [0-9\-]*[.]?/,"",s)
	gsub(/Index *included[.]?/,"",s)
	gsub(/Index[.]?/,"",s)
	gsub(/ *and index[.]?/,"",s)

	if (!( ("month" in Entry) && ("day" in Entry) ))
	{
	    if (match(s,
		      /(January|February|March|April|May|June|July|August|September|October|November|December) *[123][0-9]?, *[12][0-9][0-9][0-9]/))
	    {
		n = split(substr(s,RSTART,RLENGTH),parts," ")
		if ((n == 3) && !("year" in Entry))
		    Entry["year"] = parts[3]
		if ((n == 3) && (parts[3] == Entry["year"]))
		{
		    sub(/,$/,"",parts[2])
		    Entry["day"] = parts[2]
		    Entry["month"] = month_abbrev[tolower(parts[1])]
		}
	    }
	    else if (match(s,
		      /[123][0-9]? *(January|February|March|April|May|June|July|August|September|October|November|December), *[12][0-9][0-9][0-9]/))
	    {
		n = split(substr(s,RSTART,RLENGTH),parts," ")
		if ((n == 3) && !("year" in Entry))
		    Entry["year"] = parts[3]
		if ((n == 3) && (parts[3] == Entry["year"]))
		{
		    sub(/,$/,"",parts[2])
		    Entry["day"] = parts[1]
		    Entry["month"] = month_abbrev[tolower(parts[2])]
		}
	    }
	}

	if (s ~ /^[ ]*$/)
	    delete Entry["annote"]
	else
	    set("annote",s)
    }
}


function fix_author_c(s)
{
    ## Reduce
    ##	"Scott, Walter,|cSir,|d1771-1832"
    ## to
    ##	"Scott, Walter, Sir,|d1771-1832"
    ## and return the result
    gsub("[|$]c"," ",s)
    return (s)
}


function fix_author_d(s)
{
    ## Reduce
    ##	"Scott, Walter,|cSir,|d1771-1832"
    ## to
    ##	"Scott, Walter,|cSir"
    ## save the dates in a separate entry,
    ## and return the result
    return (subvalue(s, "d", "author-dates"))
}


function fix_author_editor_names(s)
{
    ## Reduce "Henry. McGilton and Rachel. Morgan" to
    ## "Henry McGilton and Rachel Morgan"
    ## and
    ## "Haenel, Christoph, $d 1939-" to "Haenel, Christoph"

    ## print "%% DEBUG: enter fix_author_editor_names(\"" s "\")"

    if (match(s, "^ *and +"))
	s = substr(s, RSTART + RLENGTH)

    if (match(s, "[a-z][.]"))
	s = substr(s, 1, RSTART + RLENGTH - 2) "  " fix_author_editor_names(substr(s, RSTART + RLENGTH))

    s = (s " ")
    if (match(s, "[$]d +[12][0-9][0-9][0-9]-+[^0-9]"))
	s = substr(s, 1, RSTART - 1) fix_author_editor_names(substr(s, RSTART + RLENGTH - 1))

    gsub("[ ]+[$][a-z][ ]+", " ", s)	# remove MARC subfield markers

    gsub(" +and +and +", " and ", s)
    gsub(", *[.]? *$", "", s)
    s = trim(squeeze_blanks(s))

    ## print "%% DEBUG: leave fix_author_editor_names(): [" s "]"

    return (s)
}


function fix_author_order()
{
    if ("author" in Entry)
	Entry["author"] = fix_names(Entry["author"])
}


function fix_author_q(s)
{
    ## Reduce
    ##	"Wells, H. G.|q(Herbert George),|d1866-1946"
    ## to
    ##	"Wells, H. G. (Herbert George),|d1866-1946"
    ## and return the result
    gsub("[|$]q"," ",s)
    return (s)
}


function fix_editor_order()
{
    if ("editor" in Entry)
	Entry["editor"] = fix_names(Entry["editor"])
}


function fix_ISBN(s)
{
    ## Typical values:
    ##	"0070450013 (pbk.)"
    ##	"0201134470 :, 0201134489 (pbk)",
    ##	"3540665722 (softcover : acid-free paper)"

    gsub(" *:? *acid-free,? *paper", "", s)
    gsub(" *:? *alk[.]?,? *paper,?", "", s)
    gsub(" *[(] *[Hh]c[.]? *[)]", " (hardcover)", s)
    gsub(" *[(] *[Hh]bk?[.]? *[)]", " (hardcover)", s)
    gsub(" *[(] *[Pp]bk?[.]? *[)]", " (paperback)", s)
    gsub(" *[(] *[Pp]bk?[.]? *: *alk[.],? *paper[)]", " (paperback)", s)
    gsub(" *: *, *", ", ", s)
    gsub("[(] *[)]", "", s)
    gsub("^, *", "", s)

    return (s)
}


function fix_LCCN(s,    k,key,lccns,n,parts)
{
    ## Typical values:
    ##	"QA76.76.O63 $b M68 1987"
    ##	"Z253 $b .U69 1993 090 $a Z253 $b .C57 1993 090 $a Z253 $b .C57 1993"
    gsub(" +090 +[$]a +", "; ", s)
    gsub(" +[$]b +", " ", s)

    n = split(s, parts, " *; *")
    for (k = 1; k <= n; ++k)		# construct array of unique LCCN values
	lccns[parts[k]]++

    s = ""
    for (key in lccns)			# reconstruct list of unique LCCN values
	s = (s == "") ? key : (s "; " key)

    return (s)
}


function fix_name_order(s, has_junior,has_others,has_sir,junior,n,parts,t)
{
    ## Typical values (numbers indicate parts after splitting at commas)
    #
    ## Davis, Mitchell, Jr		-> Mitchell {Davis, Jr.}	123 -> 2 {1, 3}
    ## Murphy, Harry M. , Jr		-> Harry M. {Murphy, Jr.}	123 -> 2 {1, 3}
    ## Donnelly, William J., II		-> William J. {Donnelly, II}	123 -> 2 {1, 3}
    ## Reed, J. F., III			-> J. F. {Reed, III}		123 -> 2 {1, 3}
    #
    ## Walsh, John E.			-> John E. Walsh		12 -> 2 1
    ## Walsh, John E. (John Edward)	-> John E. (John Edward) Walsh	12 -> 2 1
    ## Beethoven, Ludwig von		-> Ludwig von Beethoven		12 -> 2 1
    #
    ## von Beethoven, Ludwig		-> Ludwig {von Beethoven}	12 -> 2 {1}
    ## Auf Der Heide, Friedhelm Meyer	-> Friedhelm Meyer {Auf Der Heide} 12 -> 2 {1}
    #
    ## This case requires special handling because the word order differs from the others:
    ## Beerbohm, Max, Sir		-> Sir Max Beerbohm
    #
    ## This one requires an extra brace, because part 1 contains a non-letter:
    ## Butterworth IV, John F		-> John F. {Butterworth IV}	12 -> 2{1}
    #
    ## This one requires special handling of Jr:
    ## Lanahan, Robert J. Jr		-> Robert J. {Lanahan, Jr.}	12 -> xxxx
    ## This case suggests that we always handle Jr specially.
    #
    ## Notice that people like Guy L. Steele Jr. who omit the comma
    ## before Jr must be coded in the library catalog like this:
    ## Steele Jr., Guy L.
    ## instead of
    ## Steele, Guy L., Jr.
    ## A search of the OCLC WorldCat database for this author showed that
    ## in every case, the Jr is omitted entirely.

    gsub(/[ ,]*$/,"",s)		# remove trailing space and comma
    gsub(/[.] +[)]/,".)",s)	# "(Stuart I. )" -> "(Stuart I.)"

    t = trim(s)

    has_sir = match(t,/[, ]Sir$/) || match(t,/[, ]Count$/) || match(t,/[, ]Lord$/)
    has_junior = match(t,/[, ][JS]r[.]?$/) || match(t,/[, ][IVX]$/)

    if (has_junior)
    {
	junior = substr(t,RSTART+1,RLENGTH-1)
	if (junior ~ /^[JS][r]$/) # supply Jr./Sr. abbreviation period
	    junior = junior "."
	t = substr(t,1,RSTART-1) # remove Jr
	sub(/[, ]*$/, "", t)	# remove trailing space and comma (but NOT periods)
    }

    has_others = match(t,/[, ]et al[.]?/)
    if (has_others)
	t = delete_match(t)

    n = split(t, parts, /, */)

    parts[1] = trim(parts[1])
    if (match(parts[1]," "))
	parts[1] = "{" parts[1] "}"

    if (is_corporate_name(s))
	t = "{" trim(s)	"}"	# only brace corporate authors
    else if (n > 3)
    {
	warning("Cannot reorder a personal name with more than 3 parts: [" sample(s) "]")
	t = trim(s)
    }
    else if (n == 3)
    {
	if (has_junior)
	    t = parts[2] " {" parts[1] ", " parts[3] ", " junior "}"
	else if (has_sir)
	    t = parts[3] " " parts[2] " " parts[1]
	else
	    t = parts[2] " {" parts[1] ", " parts[3] "}"
    }
    else if (n == 2)
    {
	if (has_junior)
	    t = parts[2] " {" parts[1] ", " junior "}"
	else
	    t = parts[2] " " parts[1]
    }
    else
	t = s

    if (has_others)
	t = t " and others"
    return (t)
}


function fix_names(s, k,n,person,t)
{
    n = split(s, person, " and ")
    t = fix_name_order(person[1])
    for (k = 2; k <= n; ++k)
	t = t " and " fix_name_order(person[k])
    return (t)
}


function fix_note(s)
{
    ## Typical value: "Includes index."

    ## print "%% DEBUG: enter fix_note(\"" s "\")"
    sub("Includes index[.] *", "", s)
    ## print "%% DEBUG: leave fix_note(): s = [" s "]"

    return (s)
}


function fix_pages(s)
{
    ## Typical value: "xxv, 612 p. :"

    ## print "%% DEBUG: enter fix_pages(\"" s "\")"
    s = tolower(s)
    sub(" *: *$", "", s)
    gsub(" [ps][.] *", "", s)
    sub(" [ps]$", "", s)
    if (match(s, "^ *[ivxlcdm]+ *, *[0-9]+"))
	sub(" *, *", " + ", s)
    sub("in +various +pagings", " (various) ", s)
    sub("various +pagings", " (various) ", s)
    sub("p[.] *cm[.]", "????", s)
    sub("^ *[ps][.]? *$", "????", s)
    sub("^ *[0-9]+ *v[.]? *$", "????", s)	# "4 v." to "????"
    sub("^ *v[.]? *<[- 0-9]+> *$", "????", s)	# "v. < 1->" to "????"
    sub(";? *[0-9]+ *cm[.]?", "", s)		# discard book size

    ## print "%% DEBUG: leave fix_pages(): s = [" s "]"

    return (trim(s))
}


function fix_publisher()
{
    if (("publisher" in Entry) && (Entry["publisher"] == "Project Gutenberg"))
    {
	if (Entry_Count == 1)
	{
	    print "%-----------------------------------------------------------------------"
	    print "% Publisher abbreviations:"
	    print ""
	    print "@String{pub-PROJECT-GUTENBERG   = \"Project Gutenberg\"}"
	    print "@String{pub-PROJECT-GUTENBERG:adr = \"P.O. Box 2782,"
	    print "                                     Champaign, IL 61825-2782,"
	    print "                                     USA\"}"
	    print ""
	}
	Entry["publisher"] = "pub-PROJECT-GUTENBERG"
	if (!("address" in Entry))
	    Entry["address"] = "pub-PROJECT-GUTENBERG:adr"
    }
}


function fix_quotes(s, count,k)
{
    count = 0
    k = index(s, "\"")
    while (k > 0)
    {
	count++
	if ((count % 2) == 1)	# open quote
	    s = substr(s,1,k-1) "``" substr(s,k+1)
	else			# close quote
	    s = substr(s,1,k-1) "''" substr(s,k+1)
	k = index(s, "\"")
    }
    if ((count % 2) == 1)
	warning("Unbalanced double quotes in [" sample(s) "]")
    return (s)
}


function fix_series()
{
    if (("series" in Entry) && (Entry["series"] == "Project Gutenberg"))
    {
	if (Entry_Count == 1)
	{
	    print "%-----------------------------------------------------------------------"
	    print "% Series abbreviations:"
	    print ""
	    print "@String{ser-PROJECT-GUTENBERG = \"Project Gutenberg\"}"
	    print ""
	    print "%-----------------------------------------------------------------------"
	    print "% Bibliography entries:"
	    print ""
	}
	Entry["series"] = "ser-PROJECT-GUTENBERG"
    }
    gsub("[ ]+[$][a-z][ ]+", " ", Entry["series"])	# remove MARC subfield markers
    gsub(" *; *", "; ", Entry["series"])
    gsub(" *: *", ": ", Entry["series"])
}


function fix_subject(s,    count,k,m,n,parts)
{
    ## Typical value: "TeX (Computer file); Computerized typesetting.; Mathematics printing $x Computer programs."
    gsub(" +[$][a-z] +", "; ", s)
    gsub(" *[.];", ";", s)
    sub(" *[.] *$", "", s)
    gsub(" *, *;", ";", s)

    n = split(s, parts, Suffix["subject"])
    m = 0
    s = ""
    for (k = 1; k <= n; ++k)		# remove duplicates, but preserve original subject order
    {
	count[tolower(parts[k])]++
	if (count[tolower(parts[k])] == 1)
	    s = (s == "") ? parts[k] : (s Suffix["subject"] parts[k])
    }

    return (s)
}


function fix_subject_d(s)
{
    ## Reduce
    ##	"Shelley, Percy Bysshe,|d1792-1822."
    ## to
    ##	"Shelley, Percy Bysshe"
    ## save the dates in a separate entry,
    ## and return the result
    return (subvalue(s, "d", "subject-dates"))
}


function fix_TeX(s)
{
    gsub("[$]","\\$",s)
    gsub("[&]","\\&",s)
    gsub("[%]","\\%",s)
    return (s)
}


function fix_title_b(s, t)
{
    ## Reduce
    ## "The Early Short Fiction|bPart One /|cEdith Wharton"
    ## to
    ## "The Early Short Fiction. Part One /|cEdith Wharton"
    ## and
    ## "A Voyage to Abyssinia /|cby Father Jerome Lobo /|btranslated from the French ..."
    ## to
    ## "A Voyage to Abyssinia /|cby Father Jerome Lobo"
    ## and
    ## "A Vanished Arcadia, |bbeing some account of the Jesuits in Paraguay, 1607 to 1767..."
    ## to
    ## "A Vanished Arcadia, being some account of the Jesuits in Paraguay, 1607 to 1767..."
    ## save a note entry, and
    ## return the result

    if (match(s,"/[|$]btranslated[^|$]*"))
    {
	t = "T" substr(s,RSTART+4,RLENGTH-4)
	s = substr(s,1,RSTART-1) substr(s,RSTART+RLENGTH)
	append_suffix("note", t)
    }
    else if (match(s,"[|$]bbeing"))
	s = substr(s,1,RSTART-1) substr(s,RSTART+2)
    else
	gsub("[|$]b",". ",s)
    # print "DEBUG 1: s = [" s "]" > "/dev/stderr"
    # Remove "\$h [electronic resource]"
    gsub(" *.[$]h *\\[electronic resource\\] *", "", s)
    # print "DEBUG 2: s = [" s "]" > "/dev/stderr"
    return (s)
}


function fix_title_c(s,code, t)
{
    ## Reduce
    ##	"245 10|aAnna Karenina /|cby Leo Tolstoy/Tolstoi"
    ## to
    ##	"245 10|aAnna Karenina"
    ## save the "by ..." as a separate entry,
    ## and return the result
    if (match(s," */[|$]cby *"))
    {
	t = substr(s,RSTART+RLENGTH)
	s = substr(s,1,RSTART-1)
	append_suffix(("USMARC-" code "-c"), fix_title_b(t))
	Last_Key = "USMARC"	# NB: override value saved by append_suffix()
    }
    return (s)
}


function fix_title_l(s, t)
{
    ## Reduce
    ##	"The Augsburg Confession,|lLatin, German"
    ## to
    ##	"The Augsburg Confession"
    ## save the language value, and
    ## return the result
    if (match(s,"[|$]l[^|$]*"))
    {
	t = substr(s,RSTART+2,RLENGTH-2)
	s = substr(s,1,RSTART-1) substr(s,RSTART+RLENGTH)
	append_suffix("language", t)
    }
    return (s)
}


function fix_URL_base(s)
{
    return ((s ~ "^[a-z]+://") ? s : (URL_Base s))
}


function fix_URL_q(s, t)
{
    ## Reduce
    ##	"http://www.library.adelaide.edu.au/etext/pg/etext95/fh104s10.zip|qmidi"
    ## to
    ##	"http://www.library.adelaide.edu.au/etext/pg/etext95/fh104s10.zip"
    ## save the URL contents, and
    ## return the result
    if (match(s,"[|$]q[^|$]*"))
    {
	t = substr(s,RSTART+2,RLENGTH-2)
	s = substr(s,1,RSTART-1) substr(s,RSTART+RLENGTH)
	append_suffix("URL-contents", t)
    }
    return (s)
}


function fix_URL_z(s)
{
    ## Reduce
    ##	"http://www.library.adelaide.edu.au/etext/pg/etext95/pge0112.zip|qZIP|zVolume 1 of 28"
    ## to
    ##	"http://www.library.adelaide.edu.au/etext/pg/etext95/pge0112.zip|qZIP (Volume 1 of 28)"
    ## return the result
    if (match(s,"[|$]z[^|$]*"))
	s = substr(s,1,RSTART-1) " (" substr(s,RSTART+2,RLENGTH-2) ") " substr(s,RSTART+RLENGTH)
    return (trim(s))
}


function fix_URL(s, j,k,n,words)
{
    ## Remove a generic URL that serves no useful purpose
    gsub("ftp://uiarchive.cso.uiuc.edu/pub/etext/gutenberg/;? *", "", s)

    ## The URLs are often duplicated, so eliminate the extra ones
    n = split(s, words, "; *")
    s = words[1]
    for (k = 2; k <= n; ++k)
    {
	dups = 0
	for (j = 1; j < k; ++j)
	{
	    if (words[j] == words[k])
	        break
	}
	if (j >= k)
	    s = s "; " words[k]
    }
    return (s)
}


function fix_volume(s)
{
    ## print "%% DEBUG: enter fix_volume(\"" s "\")"
    s = trim(s)
    sub("^no[.] *", "", s)
    sub("^v(ol)?[.]? *", "", s)
    sub("[.] *$", "", s)
    ## print "%% DEBUG: leave fix_volume(): s = [" s "]"
    return (s)
}


function fix_year(s)
{
    ## print "%% DEBUG: enter fix_year(\"" s "\")"
    s = trim(s)
    sub("[.]+$", "", s)

    while (match(s, "\\[[12][0-9][0-9][0-9]\\]"))
	s = (substr(s, 1, RSTART - 1) substr(s, RSTART + 1, RLENGTH - 2) substr(s, RSTART + RLENGTH))

    sub("cop[.] *", "", s)
    if (s ~ "^c[12][0-9][0-9][0-9]$")
	s = substr(s, 2,4)
    ## print "%% DEBUG: leave fix_year(): s = [" s "]"
    return (s)
}


function initialize()
{
    ## NB: This MUST include a trailing slash
    URL_Base = "ftp://uiarchive.cso.uiuc.edu/pub/etext/gutenberg/"

    "date" | getline Current_Date_and_Time
    close("date")

    ## For the Project Gutenberg USMARC list, which is ordered by
    ## descending volume number, there are several labels which appear
    ## more than once.  This table records their frequencies for use in
    ## unique_citation_label() to generate a unique suffix letter.

    initialize_label_frequency()
    initialize_ignore_list()

    ## Multiple USMARC records of the same number are joined with these
    ## suffixes separating them:

    Suffix["abstract"] 		= ". "
    Suffix["author"] 		= " and "
    Suffix["editor"] 		= " and "
    Suffix["ISBN"] 		= ", "
    Suffix["ISSN"] 		= ", "
    Suffix["LCCN"] 		= "; "
    Suffix["language"] 		= "; "
    Suffix["note"] 		= ". "
    Suffix["publisher"] 	= " "
    Suffix["series"] 		= "; "
    Suffix["subject"] 		= "; "
    Suffix["title"] 		= " "
    Suffix["translator"] 	= "; "
    Suffix["URL"] 		= "; "
    Suffix["URL-contents"] 	= "; "
    Suffix["USMARC"] 		= "; "
    Suffix["USMARC-700"] 	= "; "
    Suffix["USMARC-740"] 	= "; "
    Suffix["USMARC-745"] 	= "; "
}


function initialize_ignore_list()
{
    ## List of words to ignore in forming citation tags.  The initial
    ## list was extracted from the bibindex badwords list, and covers
    ## a few European languages as well as English.
    Ignore["a"]         = 1
    Ignore["ab"]        = 1
    Ignore["aber"]      = 1
    Ignore["als"]       = 1
    Ignore["an"]        = 1
    Ignore["and"]       = 1
    Ignore["are"]       = 1
    Ignore["as"]        = 1
    Ignore["auf"]       = 1
    Ignore["aus"]       = 1
    Ignore["az"]        = 1
    Ignore["bei"]       = 1
    Ignore["bir"]       = 1
    Ignore["but"]       = 1
    Ignore["da"]        = 1
    Ignore["das"]       = 1
    Ignore["dat"]       = 1
    Ignore["de"]        = 1
    Ignore["dei"]       = 1
    Ignore["dem"]       = 1
    Ignore["den"]       = 1
    Ignore["der"]       = 1
    Ignore["des"]       = 1
    Ignore["det"]       = 1
    Ignore["di"]        = 1
    Ignore["die"]       = 1
    Ignore["dos"]       = 1
    Ignore["e"]         = 1
    Ignore["een"]       = 1
    Ignore["eene"]      = 1
    Ignore["egy"]       = 1
    Ignore["ei"]        = 1
    Ignore["ein"]       = 1
    Ignore["eine"]      = 1
    Ignore["einen"]     = 1
    Ignore["einer"]     = 1
    Ignore["eines"]     = 1
    Ignore["eit"]       = 1
    Ignore["el"]        = 1
    Ignore["en"]        = 1
    Ignore["er"]        = 1
    Ignore["es"]        = 1
    Ignore["et"]        = 1
    Ignore["ett"]       = 1
    Ignore["eyn"]       = 1
    Ignore["eyne"]      = 1
    Ignore["for"]       = 1
    Ignore["from"]      = 1
    Ignore["fuer"]      = 1
    Ignore["fur"]       = 1
    Ignore["gl"]        = 1
    Ignore["gli"]       = 1
    Ignore["ha"]        = 1
    Ignore["haben"]     = 1
    Ignore["had"]       = 1
    Ignore["hai"]       = 1
    Ignore["has"]       = 1
    Ignore["hat"]       = 1
    Ignore["have"]      = 1
    Ignore["he"]        = 1
    Ignore["heis"]      = 1
    Ignore["hen"]       = 1
    Ignore["hena"]      = 1
    Ignore["henas"]     = 1
    Ignore["het"]       = 1
    Ignore["hin"]       = 1
    Ignore["hinar"]     = 1
    Ignore["hinir"]     = 1
    Ignore["hinn"]      = 1
    Ignore["hith"]      = 1
    Ignore["ho"]        = 1
    Ignore["hoi"]       = 1
    Ignore["i"]         = 1
    Ignore["il"]        = 1
    Ignore["in"]        = 1
    Ignore["is"]        = 1
    Ignore["ist"]       = 1
    Ignore["ka"]        = 1
    Ignore["ke"]        = 1
    Ignore["l"]         = 1
    Ignore["la"]        = 1
    Ignore["las"]       = 1
    Ignore["le"]        = 1
    Ignore["les"]       = 1
    Ignore["lo"]        = 1
    Ignore["los"]       = 1
    Ignore["mia"]       = 1
    Ignore["mit"]       = 1
    Ignore["n"]         = 1
    Ignore["na"]        = 1
    Ignore["nji"]       = 1
    Ignore["not"]       = 1
    Ignore["o"]         = 1
    Ignore["oder"]      = 1
    Ignore["of"]        = 1
    Ignore["on"]        = 1
    Ignore["or"]        = 1
    Ignore["os"]        = 1
    Ignore["others"]    = 1
    Ignore["s"]         = 1
    Ignore["sie"]       = 1
    Ignore["sind"]      = 1
    Ignore["so"]        = 1
    Ignore["t"]         = 1
    Ignore["ta"]        = 1
    Ignore["the"]       = 1
    Ignore["to"]        = 1
    Ignore["um"]        = 1
    Ignore["uma"]       = 1
    Ignore["un"]        = 1
    Ignore["una"]       = 1
    Ignore["und"]       = 1
    Ignore["une"]       = 1
    Ignore["uno"]       = 1
    Ignore["unter"]     = 1
    Ignore["von"]       = 1
    Ignore["with"]      = 1
    Ignore["y"]         = 1
    Ignore["yr"]        = 1

#   Additional words added later
    Ignore["also"]      = 1
    Ignore["any"]       = 1
    Ignore["away"]      = 1
    Ignore["by"]        = 1
    Ignore["cum"]       = 1
    Ignore["dans"]      = 1
    Ignore["down"]      = 1
    Ignore["into"]      = 1
    Ignore["its"]       = 1
    Ignore["off"]       = 1
    Ignore["onto"]      = 1
    Ignore["out"]       = 1
    Ignore["over"]      = 1
    Ignore["sur"]       = 1
    Ignore["that"]      = 1
    Ignore["these"]     = 1
    Ignore["this"]      = 1
    Ignore["those"]     = 1
    Ignore["unto"]      = 1
    Ignore["up"]        = 1
    Ignore["via"]       = 1
    Ignore["without"]   = 1
    Ignore["zu"]        = 1
    Ignore["zum"]       = 1
    Ignore["zur"]       = 1
}


function initialize_label_frequency()
{
}


function ISBN_Digit(the_ISBN,n)
{
    return (ISSN_Digit(the_ISBN,n))	# same digit values in ISBN and ISSN data
}


function is_corporate_name(s)
{
    return (match(s,/Ltd/) || match(s,/Co[.]/) || match(s,/Company/) || \
	    match(s,/Corp/) || match(s,/Inc[.]?$/) || match(s,/Pty[.]?/) || \
	    match(s,/Cie[.]/) || match(s,/AG/))
}


function isletter(c)
{
    return (index("ABCDEFGHIJKLMNOPQRSTUVWXYZ",toupper(c)) > 0)
}


function ISSN_Digit(the_ISSN,n)
{
    return (index("0123456789X", substr(the_ISSN, n, 1)) - 1)
}


function isupper(c)
{
    return (index("ABCDEFGHIJKLMNOPQRSTUVWXYZ",c) > 0)
}


function new_entry()
{
    clear_array(Entry)
    ## Author_Editor_Key = "author"
    ## Last_Key = ""
    ## Entry["address"] = ""
    ET_AL = ""
}


function print_entry( k,key)
{
    if (!("title" in Entry))
    {
	new_entry()
	return
    }

    Entry_Count++
    k = 0
    for (key in Entry)		# count the number of saved key/value pairs
    {
	Entry[key] = fix_quotes(Entry[key])
	k++
    }

    set("entrytype","Book")	# NB: fixed type for now

    if ("author" in Entry)
	Entry["author"] = (fix_author_editor_names(Entry["author"]) ET_AL)

    if ("editor" in Entry)
	Entry["editor"] = (fix_author_editor_names(Entry["editor"]) ET_AL)

    check_for_proceedings()

    if (("entrytype" in Entry) && (k > 0))
    {
	## These two handle the special case of Project Gutenberg USMARC records
	## and must be called before any output is done
	fix_publisher()
	fix_series()

	clean_entry()

	if (entry_name() == "InProceedings")
	    print_Proceedings()	# want InProceedings to crossref a Proceedings

	print_line("@" entry_name() "{" citation_label() ",")

	if (entry_name() == "Phone")
	{
	    delete Entry["author"]
	    delete Entry["title"]
	}

	## These two must come AFTER the call to citation_label(), because
	## that function assumes the original "Last, First Middle" order
	## of names, while these functions change them to BibNet standard
	## order "First Middle Last".

	fix_author_order()
	fix_editor_order()

	if (("publisher" in Entry) && (Entry["publisher"] == "pub-PROJECT-GUTENBERG"))
	    Entry["acknowledgement"] = "ack-nhfb # \" and \" # ack-st"
	else
	    Entry["acknowledgement"] = "ack-nhfb"
	Entry["bibdate"] = Current_Date_and_Time

	## Change "not available" to something more recognizable
	if (("volume" in Entry) && (Entry["volume"] == "na"))
	    Entry["volume"] = "n/a"

	## Change note to remark.
	if ("note" in Entry)
	{
	    Entry["remark"] = Entry["note"]
	    delete Entry["note"]
	}

	if (Entry["entrytype"] == "Proceedings")
	{
	    Entry["title"]     = "{" Entry["title"] "}"
	    Entry["booktitle"] = "{" Entry["booktitle"] "}"
	}

	## After trying strict alphabetic order for the keywords, I
	## decided that it is more useful to have most of them in a
	## specific order, that corresponding to the usual order in an
	## article citation.  print_pair() will delete each entry that
	## it prints, and ignore any that don't exist, so that key/value
	## pairs are not duplicated in the subsequent output after
	## sort_keys().  The key order here is chosen to match biborder(1).

	print_pair("author")
	print_pair("editor")
	print_pair("key")
	print_pair("booktitle")
	print_pair("title")
	print_pair("crossref")
	print_pair("chapter")
	print_pair("journal")
	print_pair("volume")
	print_pair("type")
	print_pair("number")
	print_pair("howpublished")
	print_pair("institution")
	print_pair("organization")
	print_pair("publisher")
	print_pair("school")
	print_pair("address")
	print_pair("edition")
	print_pair("pages_whole")	# NB: Must precede "pages" because it
	print_pair("pages")		# checks for the existence of "pages"
	print_pair("day")
	print_pair("month")
	print_pair("year")
	print_pair("CODEN")
	print_pair("ISBN")
	print_pair("ISBN-13")
	print_pair("ISSN")
	print_pair("LCCN")
	print_pair("bibdate")
	print_pair("bibsource")
	print_pair("note")
	print_pair("price")
	print_pair("series")
	print_pair("acknowledgement")
	print_pair("subject")

	## Output any remaining key/value pairs sorted by key name
	sort_keys()

	for (k = 1; Sorted_Keys[k]; ++k)
	{
	    if (Entry[Sorted_Keys[k]] != "")
	        print_key_value(Sorted_Keys[k],Entry[Sorted_Keys[k]])
	}

	print_line("}\n")
    }

    new_entry()
}


function print_key_abbrev(key,abbrev)
{
    print_line(sprintf("  %s =%s %s,", key, \
	substr("                ",1,12-length(key)), abbrev))
}


function print_key_string(key,string)
{
    ## TeXify 8-bit accented characters
    gsub(/\252/,"\\?",string)	# what is this character?? Perhaps trademark
    gsub(/\260/,"\\?",string)	# what is this character?? Displays as Angstrom accent in willow, but as separate character
    gsub(/\341/,"\\`",string)
    gsub(/\342/,"\\'",string)
    gsub(/\343/,"\\^",string)
    gsub(/\346/,"\\?",string)	# what is this character??
    gsub(/\351/,"\\v ",string)
    gsub(/\365x/,"${}^x$",string)
    gsub(/\365X/,"${}^X$",string)
    gsub(/\350a/,"{\\\"a}",string)
    gsub(/\350e/,"{\\\"e}",string)
    gsub(/\350o/,"{\\\"o}",string)
    gsub(/\350u/,"{\\\"u}",string)

    gsub(/[.][.][.]/,"\\ldots{}",string) # convert ellipses ... to \ldots{}

    gsub(/[\;:, ]+$/,"",string)	# remove selected trailing punctuation

    print_line(sprintf("  %s = %s\"%s\",", \
	key, \
	substr("                ",1,12-length(key)), \
	trim(enquote(string))))
}


function print_key_value(key,value)
{
    if (key == "entrytype")		# already output as @type{tag, # }
	return
    else if (key == "acknowledgement")
	print_key_abbrev(key,value)
    else if ((key == "publisher") && ("publisher" in Entry) && (Entry["publisher"] ~ "pub-"))
	print_key_abbrev(key,value)
    else if ((key == "series") && ("series" in Entry) && (Entry["series"] ~ "ser-"))
	print_key_abbrev(key,value)
    else if ((key == "address") && ("address" in Entry) && (Entry["address"] ~ "pub-"))
	print_key_abbrev(key,value)
    else if (key == "month")
	print_month(value)
    else if (key == "pages_whole")
	print_key_string(("pages" in Entry) ? "pageswhole" : "pages", value)
    else
	print_key_string(key,value)
}


function print_line(line)
{
    print line
}


function print_month(month, mon)
{
    warning("print_month not yet implemented")
    return

    mon = tolower(month)
    gsub(/[^A-Za-z\055]/,"",mon)	# strip punctuation
    if (mon in month_abbrev)
	print_key_abbrev("month", month_abbrev[mon])
    else
	print_key_string("month",month)
}


function print_pair(key)
{
    if (key in Entry)
    {
	if (Entry[key] != "")
	    print_key_value(key,Entry[key])
	delete Entry[key]
    }
}


function print_Proceedings()
{
    warning("print_Proceedings() not yet implemented")
}


function sample(s)
{
    ## return up to 256 chars of s
    return ((length(s) > 256) ? (substr(s,1,256) "...") : s)
}


function set(key,value)
{
    Entry[key] = value
    Last_Key = key
}


function sort_keys( k,key,m,n)
{
    clear_array(Sorted_Keys)

    n = 0
    for (key in Entry)
    {
	n++
	Sorted_Keys[n] = key
    }
    for (k = 1; k < n; ++k)
    {
	for (m = k + 1; m <= n; ++m)
	{
	    if (tolower(Sorted_Keys[k]) > tolower(Sorted_Keys[m]))
	    {
		key = Sorted_Keys[m]
		Sorted_Keys[m] = Sorted_Keys[k]
		Sorted_Keys[k] = key
	    }
	}
    }
}


function squeeze_blanks(s)
{
    gsub(/[ \t][ \t]+/," ",s)
    return (s)
}


function subvalue(s,letter,key, t)
{
    if (match(s,"[|$]" letter "[^|$]*"))
    {
	t = substr(s,RSTART+2,RLENGTH-2)
	s = substr(s,1,RSTART-1) substr(s,RSTART+RLENGTH)
	if (key ~ "(-dates|volume)$") # change 1771-1832 to 1771--1832
	{
	    gsub("-+","--",t)
	    gsub("[.]+$","",t)
	}
	append(key, trim(t), "; ")
    }
    return (s)
}


function terminate()
{
    if (("author" in Entry) || ("title" in Entry))
	print_entry();
}


function trim(s)
{
    gsub(/^[ \t]+/,"",s)
    gsub(/[ \t]+$/,"",s)
    return (s)
}


function unique_citation_label(label)
{
    ## For the Project Gutenberg USMARC list, supply a suffix letter to
    ## make citation labels unique.
    if ((label in Label_Frequency) && \
	("publisher" in Entry) && \
	(Entry["publisher"] == "pub-PROJECT-GUTENBERG"))
    {
	if (label in Label_Count)
	    Label_Count[label]--
	else
	    Label_Count[label] = Label_Frequency[label]

	if (Label_Count[label] <= 0)
	    warning("Label_Frequency[\"" label "\"] is wrong in initialize()")
	else
	    label = label substr("abcdefghijklmnopqrstuvwxyz", Label_Count[label], 1)
    }

    ## Check for duplicate citation labels
    if (label in Label_Used)
	warning("duplicate citation label \"" label "\": add or increase Label_Frequency[] entry in initialize_label_frequency()")
    Label_Used[label]++

    return (label)
}


function unrecognized(code,s)
{
    if (s ~ "^[0-9][0-9][0-9] ")
    {
	## SUPPRESS FOR NOW: warning("unrecognized USMARC data: " s)
	s = trim(s)
	if (length(s) > 0)
	    append(("USMARC-" code), s, " ")
    }
}


function USMARC_020(s,    isbn13_pattern, k,n,parts,t)
{				# INTERNATIONAL STANDARD BOOK NUMBER (R)
    ## Typical values:
    ##	"020 $a 0070431523 (pbk.) : $c $19.95"
    ##	"020    $a 9781568812731 (alk. paper)"
    ## print "%% DEBUG: enter USMARC_020(\"" s "\")"

    isbn13_pattern = "[^0-9]978[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][^0-9]"

    if (match(s, "[$]c .*"))
    {				# cannot use subvalue here because of dollar sign in value
	t = substr(s, RSTART + 3)
	s = substr(s, 1, RSTART - 1)
	sub(" *: *$", "", s)
	gsub("[$]", "US\\$", t)
	append_suffix("price", t)
    }
    t = value(s)
    ## print "%% DEBUG: USMARC_020(): t = [" t "]"
    if (match((" " t " "), isbn13_pattern))
    {
	n = split(t, parts, " ")
	for (k = 1; k <= n; ++k)
	{
	    ## print "%% DEBUG: USMARC_020(): parts[" k "] = [" parts[k] "]"
	    if (match((" " parts[k] " "), isbn13_pattern))
	    {
		check_ISBN_13(parts[k])
	        append_suffix("ISBN-13", parts[k])
	    }
	    else
	    {
		check_ISBN(parts[k])
	        append_suffix("ISBN", parts[k])
	    }
	}
    }
    else
    {
	check_ISBN(parts[k])
	append_suffix("ISBN", t)
    }
    ## print "%% DEBUG: leave USMARC_020() subvalue = [" subvalue(s, "c", "price") "]"
}


function USMARC_021(s)
{
    USMARC_020(s)
}


function USMARC_022(s)
{				# INTERNATIONAL STANDARD SERIAL NUMBER (R)
    # Format: 020	1073-922X $y 0749-2839
    # $y means incorrect ISSN
    # $z means canceled ISSN
    gsub("[$][yz] +[0-9][0-9][0-9][0-9]-[0-9][0-9][0-9][0-9Xx]", "", s)
    s = value(s)
    check_ISSN_list(s)
    append_suffix("ISSN", s)
}


function USMARC_030(s)
{				# CODEN (Chemical Abstracts Serial Number)
    # Format: 030 	 $a JSOMAR $z JOSMAR
    # Key $a means valid CODEN, $z means canceled or invalid CODEN
    gsub("[$]z +[A-Z][A-Z][A-Z][A-Z][A-Z][A-Z0-9]", "", s)
    s = value(s)
    check_CODEN_list(s)
    append_suffix("CODEN", s)
}


function USMARC_050(s)
{				# LIBRARY OF CONGRESS CALL NUMBER (R)
    append_suffix("LCCN", value(s))
}


function USMARC_090(s)
{				# LIBRARY OF CONGRESS CALL NUMBER (R) [in University of California Melvyl system]
    USMARC_050(s)
}


function USMARC_097(s)
{				# LIBRARY OF CONGRESS CALL NUMBER (R) (not documented)
    append_suffix("LCCN", value(s))
}


function USMARC_100(s)
{				# MAIN ENTRY--PERSONAL NAME (NR)
    ## print "%% DEBUG: enter USMARC_100(): Entry[" Last_Key "] = [" Entry[Last_Key] "] s = [" s "]"
    if (match(s,"[|$]e[Ed][Dd][.] *$"))
    {
	s = substr(s,1,RSTART-1)
	Last_Key = "editor"
    }
    else
	Last_Key = "author"
    append_suffix(Last_Key, fix_author_editor_names(value(fix_author_d(fix_author_q(fix_author_c(s))))))
    ## print "%% DEBUG: leave USMARC_100(): Entry[" Last_Key "] = [" Entry[Last_Key] "] s = [" s "]"
}


function USMARC_110(s)
{				# MAIN ENTRY--CORPORATE NAME (NR)
    append_suffix("author", "{" value(s) "}")
}


function USMARC_111(s)
{				# MAIN ENTRY--MEETING NAME (NR)
    s = value(s)
    gsub(" +[$][a-z] +", " ", s)
    append_suffix("meetingname", s)
}


function USMARC_130(s)
{				# MAIN ENTRY--UNIFORM TITLE (NR)
    append_suffix("title", fix_title_l(fix_title_b(fix_title_c(value(s),"130"))))
}


function USMARC_245(s,    editors)
{				# TITLE STATEMENT (NR)
    ## Typical values:
    ##	"245 10 $a Pi-unleashed / $c Jeorg Arndt, Christoph Haenel ; translated from the German by Catriona and David Lischka."
    ##	"245 00 $a Designing data-intensive Web applications / $c Stefano Ceri ... [et al.]."
    ##	"245 00 $a Selected areas in cryptography : $b 12th international workshop, SAC 2005, Kingston, on, Canada, August 11-12, 2005, revised selected papers / $c [edited by] Bart Preneel, Stafford Tavares."

    if (match(s,"[^A-Za-z0-9]et  *al[^A-Za-z0-9]|[^A-Za-z0-9]et  *al$"))
	ET_AL = " and others"

    if (match(s, "translated from.*$"))
    {
	append_suffix("note", ("T" substr(s,RSTART+1)))
	delete_match()
    }

    if (match(s, "\\[edited by\\].*$"))
    {
	editors = substr(s,RSTART+11)
	delete_match()
	gsub(" *, *", " and ", editors)
	append_suffix("editor", editors)
    }

    append_suffix("title", fix_TeX(fix_title_b(fix_title_c(value(s),"245"))))
}


function USMARC_250(s)
{				# EDITION STATEMENT (NR)
    ## Typical value:
    ##	"250 $a 4th ed."
    s = value(s)
    sub(" *[Ee]d[.]? *", "", s)
    if (s ~ "^1st")
	return
    else if (s == "2nd")
	s = "Second"
    else if (s == "3rd")
	s = "Third"
    else if (s == "4th")
	s = "Fourth"
    else if (s == "5th")
	s = "Fifth"
    else if (s == "6th")
	s = "Sixth"
    else if (s == "7th")
	s = "Seventh"
    else if (s == "8th")
	s = "Eighth"
    else if (s == "9th")
	s = "Nineth"
    else if (s == "10th")
	s = "Tenth"
    append_suffix("edition", s)
}


function USMARC_260(s)
{				# PUBLICATION, DISTRIBUTION, ETC. (IMPRINT) (NR)
    ## append_suffix("publisher", value(subvalue(s,"c","year")))
    subvalue(s, "c", "year")
    subvalue(s, "b", "publisher")
    subvalue(s, "a", "address")
}


function USMARC_300(s)
{				# PHYSICAL DESCRIPTION (R)
    ## $a - Extent (R): The number of pages, volumes, cassettes, total playing time, etc., of the described item.
    sub("[$][b-z].*$", "", s)
    append_suffix("pages", value(s))
}


function USMARC_440(s)
{				# SERIES STATEMENT/ADDED ENTRY--TITLE (R)
    append_suffix("series", value(subvalue(s,"v","volume")))
}


function USMARC_490(s)
{				# SERIES STATEMENT/ADDED ENTRY--TITLE (R)
    ## Typical value: "490 1 $a Computers and typesetting ; $v A",
    append_suffix("series", value(subvalue(s,"v","volume")))

}


function USMARC_500(s)
{				# GENERAL NOTE (R)
    s = value(s)
    if (substr(s,length(s),1) != ".")
	s = s "."
    append_suffix("note", s)
}


function USMARC_505(s)
{				# FORMATTED CONTENTS NOTE (NR)
    append_suffix("note", value(s) ".")
}


function USMARC_520(s)
{				# SUMMARY, ETC. (R)
    append_suffix("abstract", value(s) ".")
}


function USMARC_600(s)
{				# SUBJECT ADDED ENTRY--PERSONAL NAME (R)
    append_suffix("subject", fix_subject_d(value(s)))
}


function USMARC_630(s)
{				# SUBJECT ADDED ENTRY--UNIFORM TITLE (R)
    USMARC_600(s)
}


function USMARC_650(s)
{				# SUBJECT ADDED ENTRY--TOPICAL TERM (R)
    USMARC_600(s)
}


function USMARC_651(s)
{				# SUBJECT ADDED ENTRY--GEOGRAPHIC NAME (R)
    USMARC_600(s)
}


function USMARC_690(s)
{				# Extension: appears to contain list of subject areas
    ## Typical values:
    ##	"690    $a Computer science $v Textbooks"
    ##	"690    $a Mathematics and statistics $v Textbooks"

    USMARC_600(s)
}


function USMARC_700(s)
{				# ADDED ENTRY--PERSONAL NAME (R)
    ## Typical values:
    ##	"700 1  $a Ritchie, Dennis M., $d 1941- $e joint author"
    ##	"700 1  $a Sykes, J. B. $q (John Bradbury)"

    ## print "%% DEBUG: enter USMARC_700(): Entry[" Last_Key "] = [" Entry[Last_Key] "] s = [" s "]"

    gsub("[$]e *joint *author *", " ", s)
    gsub("[$]q *", " ", s)

    if (match(s,"[|$]e[Tt][Rr][.] *$"))
	append_suffix("translator", (" and " fix_author_editor_names(value(substr(s,1,RSTART-1)))))
    else if (match(s,"[|$]eEd[.] *$"))
	append_suffix("editor", (" and " fix_author_editor_names(value(substr(s,1,RSTART-1)))))
    else
    {
	## OLD: append_suffix("USMARC-700", value(subvalue(s,"d","USMARC-700-dates")))
	append_suffix(("author" in Entry) ? "author" : "editor", (" and " fix_author_editor_names(value(s))))
    }
    ## print "%% DEBUG: leave USMARC_700(): Entry[" Last_Key "] = [" Entry[Last_Key] "] s = [" s "]"
}


function USMARC_740(s)
{				# 740 - ADDED ENTRY--UNCONTROLLED RELATED/ANALYTICAL TITLE (R)
    if ("title" in Entry)
    {
	sub("[.]", "", Entry["title"])
	Entry["title"] = Entry["title"] ": " value(s)
    }
    else
	append_suffix("USMARC-740", value(s))
}


function USMARC_745(s)
{				# 740 - ADDED ENTRY--UNCONTROLLED RELATED/ANALYTICAL TITLE (R)
    append_suffix("USMARC-745", value(s))
}


function USMARC_776(s,    lccn,k,n,parts)
{
    n = split(s, parts, " ")
    for (k = 1; k < n; ++k)
    {
	if (parts[k] == "$z")
	    append_suffix("ISBN", parts[k+1])
    }
}


function USMARC_852(s,    lccn,k,n,parts)	# http://www.loc.gov/marc/bibliographic/ecbdhold.html#mrcb852
{				# LOCATION (R)
    ## Typical values:
    ##	"852    $a SDB $b ek $h QA276.4 $i .K73 2005"
    lccn = ""
    n = split(s, parts, " ")
    for (k = 1; k < n; ++k)
    {
	if (parts[k] == "$h")
	    lccn = parts[k+1]
	else if (parts[k] == "$i")
	{
	    if (((k + 2) <= n) && (parts[k+2] ~ "^[12][0-9][0-9][0-9]"))
		lccn = (lccn " " parts[k+1] " " parts[k+2])
	    else
		lccn = (lccn " " parts[k+1])
	}
    }
    if (lccn != "")
	append_suffix("LCCN", lccn)
}


function USMARC_856(s)		# http://www.loc.gov/marc/bibliographic/ecbdhold.html#mrcb856
{				# ELECTRONIC LOCATION AND ACCESS (R)
    ## print "%% DEBUG: enter USMARC_856(\"" s "\")"
    sub(/^ *[0-9]+ +[0-9]* +/,"",s)
    ## print "%% DEBUG: clean USMARC_856(): s = [" s "]"
    s = fix_URL_base(fix_URL_z(fix_URL_q(s)))
    ## print "%% DEBUG: clean USMARC_856(): s = [" s "]"
    gsub(" *[$][a-z0-9] +", " ; ", s)
    ## print "%% DEBUG: clean USMARC_856(): s = [" s "]"
    gsub("[A-Za-z]+[ ]", " ", s)	# remove words of prose
    ## print "%% DEBUG: clean USMARC_856(): s = [" s "]"
    gsub(" +; +", "; ", s)
    gsub("; +;+", "; ", s)
    append_suffix("URL", s)
    ## print "%% DEBUG: leave USMARC_856(): s = [" s "]"
}


function value(s)
{
    ## Reduce
    ##	"100 1 |aGorky, Maxim"
    ## to
    ##	"Gorky, Maxim"
    ## by discarding the USMARC code, subcode, field letter, and trailing
    ## space, and return the result.
    sub(/^ *[0-9]+ +[0-9]* *[|$][a-z]/,"",s)
    return (trim(s))
}


function warning(message)
{
    print FILENAME ":" FNR ":%%" message >"/dev/stderr"
}
